#include <linux/config.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/param.h>
#include <linux/signal.h>
#include <linux/ioport.h>

#define	GPIO_MAJOR			250
#define GPIO_NAME			"WEB_CAMERA"
#define GPIO_INT			"GPIO_INTERRUPT"

#define SRAM_ADDRESS			0xf2020000
#define SRAM_ADDRESS_RANGE		153600
#define	IRQ_CMOS			IRQ_GPIO(17)
#define Width				240
#define Height				320
#define SIZE				76800

static void cmos_interrupt(int, void *, struct pt_regs *);
ssize_t cis_write(struct file *, const char *,size_t, loff_t *);
ssize_t cis_read(struct file *, char *, size_t,loff_t *);
static int cis_open(struct inode *, struct file *);
static int cis_release(struct inode *, struct file *);

static pid_t id;
unsigned int *pSRAM;
unsigned short ycbcr422[SIZE], RGB[SIZE];

static struct file_operations cis_fops = {
	write:			cis_write,
	read:			cis_read,
	open:			cis_open,
	release:		cis_release,
};

int init_module(viod){
	int result;
	
	result = register_chrdev(GPIO_MAJOR, GPIO_INT, &cis_fops);

	if (!result){
		if(!check_region(SRAM_ADDRESS, SRAM_ADDRESS_RANGE)){
			request_region(SRAM_ADDRESS, SRAM_ADDRESS_RANGE, GPIO_NAME);
	printk(/*KERN_NOTICE*/ "I/O Region Checking ok...\n");
	printk(/*KERN_NOTICE*/ "I/O Region Registration to Kernel ok...\n");
	printk(/*KERN_NOTICE*/ "Authorized Major Number[=%d] of Driver \"%s\" is allocated successfully...\n",GPIO_MAJOR, GPIO_NAME);
		}
	else	printk(/*KERN_ERR*/ "Region Checking of Driver \"%s\" is Failed\n",GPIO_NAME);
	}
	else if (result < 0){
	printk(/*KERN_ERR*/ "Can't get any Major Number in Driver  \"%s\" is Failed\n",GPIO_NAME);

	return result;
	}
	return 0;
}

void cleanup_module(void){
	relaese_region(SRAM_ADDRESS, SRAM_ADDRESS_RANGE);

	if (!unregister_chrdev(GPIO_MAJOR, GPIO_INT)){
		pinrtk(/*KERN_NOTICE*/ "I/O Region Release ok...\n");
		printk(/*KERN_NOTICE*/ "Driver \"%s\" CLEAN_UP SUCCESS~!\n", GPIO_NAME);
	}
	else	printk(/*KERN_ERR*/ "Driver \"%s\" CLEAN_UP FAILURE..\n",GPIO_NAME);
}

static void cmos_interrupt(int irq, void *dev_id, struct pt_regs * regs){
	int i,j,n,R,G,B,Y1,Y2,Cb1,Cr1,Cr2;

	disable_irq(IRQ_CMOS);

	pSRAM = (unsigned int *)SRAM_ADDRESS;
	memcpy(ycbcr422, pSRAM, SIZE * 2);

	n = Cr1 = Cr2 = 0;
	for (j = 0; j < Width; j++){
		for(i = (Height - 1); i >= 0; i -= 2){
			Y1 = (ycbcr422[n] & 0xff00);
			Cb1 = (ycbcr422[n++] & 0xff) -128;
			Y2 = (ycbcr422[n] & 0xff00);
			Cr2 = (ycbcr422[n++] & 0xff) -128;

			R = Y1 + 351 * Cr1;
			G = Y1 - 179 * Cr1 - 86 * Cb1;
			B = Y1 + 443 * Cb1;

			R = (R > 0xff00) ? 0xff00 : ( R < 0) ? 0x0 : R; 
			G = (G > 0xff00) ? 0xff00 : ( G < 0) ? 0x0 : G; 
			B = (B > 0xff00) ? 0xff00 : ( B < 0) ? 0x0 : B;

			RGB[i * Width + j] = ( R & 0xf800) | (( G >> 5) &  0x7e0) | ((B >> 11) & 0x1f);

			R = Y2 + 351 *Cr2;
			G = Y2 - 179 * Cr2 - 86 * Cb1;
			B = Y2 + 443 * Cb1;

			R = (R > 0xff00) ? 0xff00 : ( R < 0) ? 0x0 : R; 
			G = (G > 0xff00) ? 0xff00 : ( G < 0) ? 0x0 : G; 
			B = (B > 0xff00) ? 0xff00 : ( B < 0) ? 0x0 : B; 

			RGB[(i-1) * Width + j] = (R & 0x8f00) | ((G >> 5) & 0x7e0) | ((B >> 11) & 0x1f);
		}
	}
	kill_proc(id, SIGUSR2 ,1);
}

static int cis_open(struct inode *inode, struct file *filp){

	int res;
	unsigned int gafr;
	
	CKEN |= CKEN11_USB;

	gafr = 0x4000;
	GAFR0_L |= gafr;

	GPDR(IRQ_TO_GPIO_2_80(IRQ_CMOS)) &= ~GPIO_bit(IRQ_TO_GPIO_2_80(IRQ_CMOS));
	set_GPIO_IRQ_edge(IRQ_TO_GPIO_2_80(IRQ_CMOS), GPIO_FALLING_EDGE);

	res = request_irq(IRQ_CMOS, &cmos_interrupt, SA_INTERRUPT, "CMOS",NULL);

	if(res < 0) printk(/*KERN_ERR*/ " %s: Request for IRQ &d failed\n",__FUNCTION__,IRQ_CMOS);
	else	enable_irq(IRQ_CMOS);

	pSRAM = (unsigned int *)SRAM_ADDRESS;

	MOD_INC_USE_COUNT;
	
	return 0;
}

ssize_t cis_write(struct file *inode, const char *gdata, size_t length, loff_t *off_what){
	get_user(id, (int *)gdata);
	enable_irq(IRQ_CMOS);

	return length;
}

ssize_t cis_read(struct file *flip, char *buf, size_t count, loff_t *l){
	copy_to_user(buf, RGB, SIZE * 2);
	return count;
}

static int cis_release(struct inode *inode, struct file *filp){

	free_irq(IRQ_CMOS,NULL);
	disable_irq(IRQ_CMOS);
	CKEN =CKEN & ~CKEN11_USB;
	MOD_DEC_USE_COUNT;
	return 0;
}

